@@ -1,7 +1,9 @@
 use std::{num::NonZeroU32, ops::Range};
 
-use euclid::{Rect, Size2D};
-use wgpu::{FilterMode, TextureFormat};
+use figures::{Pixels, Size, SizedRect};
+use wgpu::{
+    FilterMode, MultisampleState, SurfaceFrame, TextureAspect, TextureFormat, TextureViewDescriptor,
+};
 
 use crate::{
     binding::{Bind, BindingGroup, BindingGroupLayout},
@@ -13,7 +15,6 @@ use crate::{
     frame::Frame,
     pipeline::{AbstractPipeline, Blending},
     sampler::Sampler,
-    swapchain::SwapChain,
     texture::Texture,
     transform::ScreenSpace,
     vertex::VertexLayout,
@@ -26,12 +27,15 @@ pub trait Draw {
 #[derive(Debug)]
 pub struct Renderer {
     pub device: Device,
+    /// Enables MSAA for values > 1.
+    pub(crate) sample_count: u32,
 }
 
 impl Renderer {
     pub async fn for_surface(
         surface: wgpu::Surface,
         instance: &wgpu::Instance,
+        sample_count: u32,
     ) -> Result<Self, Error> {
         let adapter = instance
             .request_adapter(&wgpu::RequestAdapterOptions {
@@ -43,48 +47,70 @@ impl Renderer {
 
         Ok(Self {
             device: Device::for_surface(surface, &adapter).await?,
+            sample_count,
         })
     }
 
-    pub async fn offscreen(adapter: &wgpu::Adapter) -> Result<Self, Error> {
+    pub async fn offscreen(adapter: &wgpu::Adapter, sample_count: u32) -> Result<Self, Error> {
         Ok(Self {
-            device: Device::offscreen(&adapter).await?,
+            device: Device::offscreen(adapter).await?,
+            sample_count,
         })
     }
 
-    pub fn swap_chain<PresentMode: Into<wgpu::PresentMode>>(
-        &self,
-        size: Size2D<u32, ScreenSpace>,
+    pub const fn sample_count(&self) -> u32 {
+        self.sample_count
+    }
+
+    pub fn configure<PresentMode: Into<wgpu::PresentMode>>(
+        &mut self,
+        size: Size<u32, ScreenSpace>,
         mode: PresentMode,
         format: TextureFormat,
-    ) -> SwapChain {
-        SwapChain {
-            depth: self.device.create_zbuffer(size),
-            wgpu: self.device.create_swap_chain(size, mode, format),
-            size,
-            format,
-        }
+    ) {
+        self.device.configure(size, mode, format)
+    }
+
+    pub fn current_frame(&self) -> Result<RenderFrame, wgpu::SurfaceError> {
+        let surface = self.device.surface.as_ref().unwrap();
+        let surface_frame = surface.get_current_frame()?;
+        let view = surface_frame
+            .output
+            .texture
+            .create_view(&TextureViewDescriptor::default());
+        Ok(RenderFrame {
+            wgpu: surface_frame,
+            view,
+            depth: self
+                .device
+                .create_zbuffer(self.device.size(), self.sample_count),
+            size: self.device.size(),
+        })
     }
 
     pub fn texture(
         &self,
-        size: Size2D<u32, ScreenSpace>,
+        size: Size<u32, ScreenSpace>,
         format: wgpu::TextureFormat,
-        usage: wgpu::TextureUsage,
+        usage: wgpu::TextureUsages,
+        multisampled: bool,
     ) -> Texture {
-        self.device.create_texture(size, format, usage)
+        let sample_count = if multisampled { self.sample_count } else { 1 };
+        self.device
+            .create_texture(size, format, usage, sample_count)
     }
 
     pub fn framebuffer(
         &self,
-        size: Size2D<u32, ScreenSpace>,
+        size: Size<u32, ScreenSpace>,
         format: wgpu::TextureFormat,
     ) -> Framebuffer {
-        self.device.create_framebuffer(size, format)
+        self.device
+            .create_framebuffer(size, format, self.sample_count)
     }
 
-    pub fn zbuffer(&self, size: Size2D<u32, ScreenSpace>) -> DepthBuffer {
-        self.device.create_zbuffer(size)
+    pub fn zbuffer(&self, size: Size<u32, ScreenSpace>) -> DepthBuffer {
+        self.device.create_zbuffer(size, self.sample_count)
     }
 
     pub fn vertex_buffer<T: bytemuck::Pod>(&self, verts: &[T]) -> VertexBuffer
@@ -120,8 +146,19 @@ impl Renderer {
         let fs = self.device.create_shader(desc.fragment_shader);
 
         T::setup(
-            self.device
-                .create_pipeline(pip_layout, vertex_layout, blending, &vs, &fs, format),
+            self.device.create_pipeline(
+                pip_layout,
+                vertex_layout,
+                blending,
+                &vs,
+                &fs,
+                format,
+                MultisampleState {
+                    count: self.sample_count,
+                    mask: !0,
+                    alpha_to_coverage_enabled: false,
+                },
+            ),
             &self.device,
         )
     }
@@ -136,7 +173,7 @@ impl Renderer {
         let gpu_buffer = self.device.wgpu.create_buffer(&wgpu::BufferDescriptor {
             label: None,
             size: bytesize as u64,
-            usage: wgpu::BufferUsage::MAP_READ | wgpu::BufferUsage::COPY_DST,
+            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
             mapped_at_creation: false,
         });
 
@@ -145,6 +182,7 @@ impl Renderer {
                 texture: &fb.texture.wgpu,
                 mip_level: 0,
                 origin: wgpu::Origin3d { x: 0, y: 0, z: 0 },
+                aspect: TextureAspect::All,
             },
             wgpu::ImageCopyBuffer {
                 buffer: &gpu_buffer,
@@ -213,12 +251,12 @@ pub enum Op<'a, T> {
     Transfer {
         f: &'a dyn Canvas<Color = T>,
         buf: &'a [T],
-        rect: Rect<i32, ScreenSpace>,
+        rect: SizedRect<i32, ScreenSpace>,
     },
     Blit(
         &'a dyn Canvas<Color = T>,
-        Rect<u32, ScreenSpace>,
-        Rect<u32, ScreenSpace>,
+        SizedRect<u32, ScreenSpace>,
+        SizedRect<u32, ScreenSpace>,
     ),
 }
 
@@ -248,6 +286,7 @@ pub trait RenderPassExt<'a> {
     fn begin(
         encoder: &'a mut wgpu::CommandEncoder,
         view: &'a wgpu::TextureView,
+        resolve_target: Option<&'a wgpu::TextureView>,
         depth: &'a wgpu::TextureView,
         op: PassOp,
     ) -> Self;
@@ -270,14 +309,15 @@ impl<'a> RenderPassExt<'a> for wgpu::RenderPass<'a> {
     fn begin(
         encoder: &'a mut wgpu::CommandEncoder,
         view: &'a wgpu::TextureView,
+        resolve_target: Option<&'a wgpu::TextureView>,
         depth: &'a wgpu::TextureView,
         op: PassOp,
     ) -> Self {
         encoder.begin_render_pass(&wgpu::RenderPassDescriptor {
             label: None,
             color_attachments: &[wgpu::RenderPassColorAttachment {
-                view: &view,
-                resolve_target: None,
+                view,
+                resolve_target,
                 ops: wgpu::Operations {
                     load: op.to_wgpu(),
                     store: true,
@@ -358,3 +398,20 @@ pub trait RenderTarget {
     /// Depth component.
     fn zdepth_target(&self) -> &wgpu::TextureView;
 }
+
+pub struct RenderFrame {
+    pub view: wgpu::TextureView,
+    pub wgpu: SurfaceFrame,
+    pub depth: DepthBuffer,
+    pub size: Size<u32, Pixels>,
+}
+
+impl RenderTarget for RenderFrame {
+    fn color_target(&self) -> &wgpu::TextureView {
+        &self.view
+    }
+
+    fn zdepth_target(&self) -> &wgpu::TextureView {
+        &self.depth.texture.view
+    }
+}
